import pygad
import random

num_generations = 100
sol_per_pop = 10
num_parents_mating = 5

# TODO Verify that each entry in 'solutions_fitness' and 'best_solutions_fitness' has values equal to the number of objectives.

def number_saved_solutions(keep_elitism=1, 
                           keep_parents=-1,
                           mutation_type="random",
                           mutation_percent_genes="default",
                           parent_selection_type='sss',
                           multi_objective=False,
                           fitness_batch_size=None,
                           save_solutions=False,
                           save_best_solutions=False):

    def fitness_func_no_batch_single(ga, solution, idx):
        return random.random()

    def fitness_func_no_batch_multi(ga_instance, solution, solution_idx):
        return [random.random(), random.random()]

    def fitness_func_batch_single(ga_instance, solution, solution_idx):
        f = []
        for sol in solution:
            f.append(random.random())
        return f

    def fitness_func_batch_multi(ga_instance, solution, solution_idx):
        f = []
        for sol in solution:
            f.append([random.random(), random.random()])
        return f

    if fitness_batch_size is None or (type(fitness_batch_size) in pygad.GA.supported_int_types and fitness_batch_size == 1):
        if multi_objective == True:
            fitness_func = fitness_func_no_batch_multi
        else:
            fitness_func = fitness_func_no_batch_single
    elif (type(fitness_batch_size) in pygad.GA.supported_int_types and fitness_batch_size > 1):
        if multi_objective == True:
            fitness_func = fitness_func_batch_multi
        else:
            fitness_func = fitness_func_batch_single

    ga_optimizer = pygad.GA(num_generations=num_generations,
                            sol_per_pop=sol_per_pop,
                            num_genes=6,
                            num_parents_mating=num_parents_mating,
                            fitness_func=fitness_func,
                            mutation_type=mutation_type,
                            parent_selection_type=parent_selection_type,
                            mutation_percent_genes=mutation_percent_genes,
                            keep_elitism=keep_elitism,
                            keep_parents=keep_parents,
                            suppress_warnings=True,
                            fitness_batch_size=fitness_batch_size,
                            save_best_solutions=save_best_solutions,
                            save_solutions=save_solutions)

    ga_optimizer.run()

    if save_solutions == True:
        expected_num_solutions = sol_per_pop + num_generations * sol_per_pop
    else:
        expected_num_solutions = 0

    if save_best_solutions == True:
        expected_num_best_solutions = 1 + num_generations
    else:
        expected_num_best_solutions = 0

    print(f"Expected number of solutions is {expected_num_solutions}.")
    print(f"Actual number of solutions is {len(ga_optimizer.solutions)}.")
    print(f"Expected number of best solutions is {expected_num_best_solutions}.")
    print(f"Actual number of best solutions is {len(ga_optimizer.best_solutions)}.")
    return expected_num_solutions, len(ga_optimizer.solutions), len(ga_optimizer.solutions_fitness), expected_num_best_solutions, len(ga_optimizer.best_solutions)

#### Single Objective
def test_save_solutions_default_keep():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions()
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions()
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_save_solutions():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(save_solutions=True,
                                                                                                                                            save_best_solutions=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_elitism():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=0)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_elitism_save_solutions():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_parents():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_parents_save_solutions():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_save_solutions():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_adaptive_mutation():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5])
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_adaptive_mutation_save_solutions():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_default_adaptive_mutation():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5])
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_adaptive_mutation():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5])
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_adaptive_mutation_save_solutions():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions


#### Multi Objective
def test_save_solutions_default_keep_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_save_solutions_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_elitism_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_elitism_save_solutions_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_parents_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_parents_save_solutions_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_save_solutions_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_adaptive_mutation_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_adaptive_mutation_save_solutions_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_default_adaptive_mutation_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_adaptive_mutation_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_adaptive_mutation_save_solutions_multi_objective():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions


#### Multi Objective NSGA-II Parent Selection
def test_save_solutions_default_keep_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_save_solutions_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_elitism_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_elitism_save_solutions_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_parents_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_parents_save_solutions_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_save_solutions_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_adaptive_mutation_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_adaptive_mutation_save_solutions_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_default_adaptive_mutation_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_adaptive_mutation_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_adaptive_mutation_save_solutions_multi_objective_nsga2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2')
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

######## Batch Fitness
#### Single Objective
def test_save_solutions_no_keep_batch_1():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(fitness_batch_size=1)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_batch_2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(fitness_batch_size=2)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_batch_3():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(fitness_batch_size=3)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_batch_5():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(fitness_batch_size=5)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_batch_6():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(fitness_batch_size=6)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_batch_7():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(fitness_batch_size=7)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_batch_8():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(fitness_batch_size=8)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_batch_9():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(fitness_batch_size=9)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_batch_10():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(fitness_batch_size=10)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions


def test_save_solutions_no_keep_save_solutions_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_elitism_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_elitism_save_solutions_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_parents_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_parents_save_solutions_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_save_solutions_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_adaptive_mutation_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_adaptive_mutation_save_solutions_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_default_adaptive_mutation_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_adaptive_mutation_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_adaptive_mutation_save_solutions_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions


#### Multi Objective
def test_save_solutions_no_keep_multi_objective_batch_1():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            fitness_batch_size=1)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_batch_2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            fitness_batch_size=2)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_batch_3():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            fitness_batch_size=3)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_batch_5():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            fitness_batch_size=5)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_batch_6():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            fitness_batch_size=6)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_batch_7():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            fitness_batch_size=7)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_batch_8():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            fitness_batch_size=8)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_batch_9():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            fitness_batch_size=9)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_batch_10():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            fitness_batch_size=10)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_save_solutions_multi_objective_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_elitism_multi_objective_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            multi_objective=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_elitism_save_solutions_multi_objective_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_parents_multi_objective_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            multi_objective=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_parents_save_solutions_multi_objective_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            multi_objective=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_multi_objective_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            multi_objective=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_save_solutions_multi_objective_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_adaptive_mutation_multi_objective_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            multi_objective=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_adaptive_mutation_save_solutions_multi_objective_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_default_adaptive_mutation_multi_objective_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            multi_objective=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_adaptive_mutation_multi_objective_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            multi_objective=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_adaptive_mutation_save_solutions_multi_objective_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions


#### Multi Objective NSGA-II Parent Selection
def test_save_solutions_no_keep_multi_objective_nsga2_batch_1():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=1)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_nsga2_batch_2():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=2)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_nsga2_batch_3():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=3)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_nsga2_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_nsga2_batch_5():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=5)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_nsga2_batch_6():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=6)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_nsga2_batch_7():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=7)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_nsga2_batch_8():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=8)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_nsga2_batch_9():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=9)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_multi_objective_nsga2_batch_10():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=10)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_save_solutions_multi_objective_nsga2_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_elitism_multi_objective_nsga2_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_elitism_save_solutions_multi_objective_nsga2_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_parents_multi_objective_nsga2_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_keep_parents_save_solutions_multi_objective_nsga2_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_multi_objective_nsga2_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_save_solutions_multi_objective_nsga2_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_adaptive_mutation_multi_objective_nsga2_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_no_keep_adaptive_mutation_save_solutions_multi_objective_nsga2_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=0, 
                                                                                                                                            keep_parents=0,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_default_adaptive_mutation_multi_objective_nsga2_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_adaptive_mutation_multi_objective_nsga2_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions

def test_save_solutions_both_keep_adaptive_mutation_save_solutions_multi_objective_nsga2_batch_4():
    expected_solutions, actual_solutions, actual_solutions_fitness, expected_best_solutions, actual_best_solutions = number_saved_solutions(keep_elitism=3, 
                                                                                                                                            keep_parents=4,
                                                                                                                                            mutation_type="adaptive",
                                                                                                                                            mutation_percent_genes=[10, 5],
                                                                                                                                            save_solutions=True,
                                                                                                                                            save_best_solutions=True,
                                                                                                                                            multi_objective=True,
                                                                                                                                            parent_selection_type='nsga2',
                                                                                                                                            fitness_batch_size=4)
    assert expected_solutions == actual_solutions
    assert expected_solutions == actual_solutions_fitness
    assert expected_best_solutions == actual_best_solutions




if __name__ == "__main__":
    #### Single Objective
    print()
    test_save_solutions_default_keep()
    print()
    test_save_solutions_no_keep()
    print()
    test_save_solutions_no_keep_save_solutions()
    print()
    test_save_solutions_keep_elitism()
    print()
    test_save_solutions_keep_elitism_save_solutions()
    print()
    test_save_solutions_keep_parents()
    print()
    test_save_solutions_keep_parents_save_solutions()
    print()
    test_save_solutions_both_keep()
    print()
    test_save_solutions_both_keep_save_solutions()
    print()
    test_save_solutions_no_keep_adaptive_mutation()
    print()
    test_save_solutions_no_keep_adaptive_mutation_save_solutions()
    print()
    test_save_solutions_default_adaptive_mutation()
    print()
    test_save_solutions_both_keep_adaptive_mutation()
    print()
    test_save_solutions_both_keep_adaptive_mutation_save_solutions()
    print()

    #### Multi-Objective
    print()
    test_save_solutions_default_keep_multi_objective()
    print()
    test_save_solutions_no_keep_multi_objective()
    print()
    test_save_solutions_no_keep_save_solutions_multi_objective()
    print()
    test_save_solutions_keep_elitism_multi_objective()
    print()
    test_save_solutions_keep_elitism_save_solutions_multi_objective()
    print()
    test_save_solutions_keep_parents_multi_objective()
    print()
    test_save_solutions_keep_parents_save_solutions_multi_objective()
    print()
    test_save_solutions_both_keep_multi_objective()
    print()
    test_save_solutions_both_keep_save_solutions_multi_objective()
    print()
    test_save_solutions_no_keep_adaptive_mutation_multi_objective()
    print()
    test_save_solutions_no_keep_adaptive_mutation_save_solutions_multi_objective()
    print()
    test_save_solutions_default_adaptive_mutation_multi_objective()
    print()
    test_save_solutions_both_keep_adaptive_mutation_multi_objective()
    print()
    test_save_solutions_both_keep_adaptive_mutation_save_solutions_multi_objective()
    print()


    #### Multi-Objective NSGA-II Parent Selection
    print()
    test_save_solutions_default_keep_multi_objective_nsga2()
    print()
    test_save_solutions_no_keep_multi_objective_nsga2()
    print()
    test_save_solutions_no_keep_save_solutions_multi_objective_nsga2()
    print()
    test_save_solutions_keep_elitism_multi_objective_nsga2()
    print()
    test_save_solutions_keep_elitism_save_solutions_multi_objective_nsga2()
    print()
    test_save_solutions_keep_parents_multi_objective_nsga2()
    print()
    test_save_solutions_keep_parents_save_solutions_multi_objective_nsga2()
    print()
    test_save_solutions_both_keep_multi_objective_nsga2()
    print()
    test_save_solutions_both_keep_save_solutions_multi_objective_nsga2()
    print()
    test_save_solutions_no_keep_adaptive_mutation_multi_objective_nsga2()
    print()
    test_save_solutions_no_keep_adaptive_mutation_save_solutions_multi_objective_nsga2()
    print()
    test_save_solutions_default_adaptive_mutation_multi_objective_nsga2()
    print()
    test_save_solutions_both_keep_adaptive_mutation_multi_objective_nsga2()
    print()
    test_save_solutions_both_keep_adaptive_mutation_save_solutions_multi_objective_nsga2()
    print()

    ######## Batch Fitness Calculation
    #### Single Objective
    print()
    test_save_solutions_no_keep_batch_1()
    print()
    test_save_solutions_no_keep_batch_2()
    print()
    test_save_solutions_no_keep_batch_3()
    print()
    test_save_solutions_no_keep_batch_4()
    print()
    test_save_solutions_no_keep_batch_5()
    print()
    test_save_solutions_no_keep_batch_6()
    print()
    test_save_solutions_no_keep_batch_7()
    print()
    test_save_solutions_no_keep_batch_8()
    print()
    test_save_solutions_no_keep_batch_9()
    print()
    test_save_solutions_no_keep_batch_10()
    print()
    test_save_solutions_no_keep_save_solutions_batch_4()
    print()
    test_save_solutions_keep_elitism_batch_4()
    print()
    test_save_solutions_keep_elitism_save_solutions_batch_4()
    print()
    test_save_solutions_keep_parents_batch_4()
    print()
    test_save_solutions_keep_parents_save_solutions_batch_4()
    print()
    test_save_solutions_both_keep_batch_4()
    print()
    test_save_solutions_both_keep_save_solutions_batch_4()
    print()
    test_save_solutions_no_keep_adaptive_mutation_batch_4()
    print()
    test_save_solutions_no_keep_adaptive_mutation_save_solutions_batch_4()
    print()
    test_save_solutions_default_adaptive_mutation_batch_4()
    print()
    test_save_solutions_both_keep_adaptive_mutation_batch_4()
    print()
    test_save_solutions_both_keep_adaptive_mutation_save_solutions_batch_4()
    print()

    #### Multi-Objective
    print()
    test_save_solutions_no_keep_multi_objective_batch_1()
    print()
    test_save_solutions_no_keep_multi_objective_batch_2()
    print()
    test_save_solutions_no_keep_multi_objective_batch_3()
    print()
    test_save_solutions_no_keep_multi_objective_batch_4()
    print()
    test_save_solutions_no_keep_multi_objective_batch_5()
    print()
    test_save_solutions_no_keep_multi_objective_batch_6()
    print()
    test_save_solutions_no_keep_multi_objective_batch_7()
    print()
    test_save_solutions_no_keep_multi_objective_batch_8()
    print()
    test_save_solutions_no_keep_multi_objective_batch_9()
    print()
    test_save_solutions_no_keep_multi_objective_batch_10()
    print()
    test_save_solutions_no_keep_save_solutions_multi_objective_batch_4()
    print()
    test_save_solutions_keep_elitism_multi_objective_batch_4()
    print()
    test_save_solutions_keep_elitism_save_solutions_multi_objective_batch_4()
    print()
    test_save_solutions_keep_parents_multi_objective_batch_4()
    print()
    test_save_solutions_keep_parents_save_solutions_multi_objective_batch_4()
    print()
    test_save_solutions_both_keep_multi_objective_batch_4()
    print()
    test_save_solutions_both_keep_save_solutions_multi_objective_batch_4()
    print()
    test_save_solutions_no_keep_adaptive_mutation_multi_objective_batch_4()
    print()
    test_save_solutions_no_keep_adaptive_mutation_save_solutions_multi_objective_batch_4()
    print()
    test_save_solutions_default_adaptive_mutation_multi_objective_batch_4()
    print()
    test_save_solutions_both_keep_adaptive_mutation_multi_objective_batch_4()
    print()
    test_save_solutions_both_keep_adaptive_mutation_save_solutions_multi_objective_batch_4()
    print()


    #### Multi-Objective NSGA-II Parent Selection
    print()
    test_save_solutions_no_keep_multi_objective_nsga2_batch_1()
    print()
    test_save_solutions_no_keep_multi_objective_nsga2_batch_2()
    print()
    test_save_solutions_no_keep_multi_objective_nsga2_batch_3()
    print()
    test_save_solutions_no_keep_multi_objective_nsga2_batch_4()
    print()
    test_save_solutions_no_keep_multi_objective_nsga2_batch_5()
    print()
    test_save_solutions_no_keep_multi_objective_nsga2_batch_6()
    print()
    test_save_solutions_no_keep_multi_objective_nsga2_batch_7()
    print()
    test_save_solutions_no_keep_multi_objective_nsga2_batch_8()
    print()
    test_save_solutions_no_keep_multi_objective_nsga2_batch_9()
    print()
    test_save_solutions_no_keep_multi_objective_nsga2_batch_10()
    print()
    test_save_solutions_no_keep_save_solutions_multi_objective_nsga2_batch_4()
    print()
    test_save_solutions_keep_elitism_multi_objective_nsga2_batch_4()
    print()
    test_save_solutions_keep_elitism_save_solutions_multi_objective_nsga2_batch_4()
    print()
    test_save_solutions_keep_parents_multi_objective_nsga2_batch_4()
    print()
    test_save_solutions_keep_parents_save_solutions_multi_objective_nsga2_batch_4()
    print()
    test_save_solutions_both_keep_multi_objective_nsga2_batch_4()
    print()
    test_save_solutions_both_keep_save_solutions_multi_objective_nsga2_batch_4()
    print()
    test_save_solutions_no_keep_adaptive_mutation_multi_objective_nsga2_batch_4()
    print()
    test_save_solutions_no_keep_adaptive_mutation_save_solutions_multi_objective_nsga2_batch_4()
    print()
    test_save_solutions_default_adaptive_mutation_multi_objective_nsga2_batch_4()
    print()
    test_save_solutions_both_keep_adaptive_mutation_multi_objective_nsga2_batch_4()
    print()
    test_save_solutions_both_keep_adaptive_mutation_save_solutions_multi_objective_nsga2_batch_4()
    print()



"""
The pygad.utils.mutation module has all the built-in mutation operators.
"""

import numpy
import random

import pygad
import concurrent.futures

import warnings

class Mutation:

    def __init__(self):
        pass

    def random_mutation(self, offspring):

        """
        Applies the random mutation which changes the values of a number of genes randomly.
        The random value is selected either using the 'gene_space' parameter or the 2 parameters 'random_mutation_min_val' and 'random_mutation_max_val'.

        It accepts:
            -offspring: The offspring to mutate.

        It returns an array of the mutated offspring.
        """

        # If the mutation values are selected from the mutation space, the attribute 'gene_space' is not None. Otherwise, it is None.
        # When the 'mutation_probability' parameter exists (i.e. not None), then it is used in the mutation. Otherwise, the 'mutation_num_genes' parameter is used.

        if self.mutation_probability is None:
            # When the 'mutation_probability' parameter does not exist (i.e. None), then the parameter 'mutation_num_genes' is used in the mutation.
            if not (self.gene_space is None):
                # When the attribute 'gene_space' exists (i.e. not None), the mutation values are selected randomly from the space of values of each gene.
                offspring = self.mutation_by_space(offspring)
            else:
                offspring = self.mutation_randomly(offspring)
        else:
            # When the 'mutation_probability' parameter exists (i.e. not None), then it is used in the mutation.
            if not (self.gene_space is None):
                # When the attribute 'gene_space' does not exist (i.e. None), the mutation values are selected randomly based on the continuous range specified by the 2 attributes 'random_mutation_min_val' and 'random_mutation_max_val'.
                offspring = self.mutation_probs_by_space(offspring)
            else:
                offspring = self.mutation_probs_randomly(offspring)

        return offspring

    def mutation_by_space(self, offspring):

        """
        Applies the mutation using the gene_space parameter.
        It accepts:
            -offspring: The offspring to mutate.
        It returns an array of the mutated offspring using the mutation space.
        """

        # For each offspring, a value from the gene space is selected randomly and assigned to the selected mutated gene.
        for offspring_idx in range(offspring.shape[0]):
            mutation_indices = numpy.array(random.sample(range(0, self.num_genes), self.mutation_num_genes))
            for gene_idx in mutation_indices:

                value_from_space = self.mutation_process_gene_value(solution=offspring[offspring_idx],
                                                                    gene_idx=gene_idx,
                                                                    sample_size=self.sample_size)

                # Before assigning the selected value from the space to the gene, change its data type and round it.
                if self.gene_type_single == True:
                    if not self.gene_type[1] is None:
                        offspring[offspring_idx, gene_idx] = numpy.round(self.gene_type[0](value_from_space),
                                                                         self.gene_type[1])
                    else:
                        offspring[offspring_idx, gene_idx] = self.gene_type[0](value_from_space)
                else:
                    if not self.gene_type[gene_idx][1] is None:
                        offspring[offspring_idx, gene_idx] = numpy.round(self.gene_type[gene_idx][0](value_from_space),
                                                                         self.gene_type[gene_idx][1])
                    else:
                        offspring[offspring_idx, gene_idx] = self.gene_type[gene_idx][0](value_from_space)

                if self.allow_duplicate_genes == False:
                    offspring[offspring_idx], _, _ = self.solve_duplicate_genes_by_space(solution=offspring[offspring_idx],
                                                                                         gene_type=self.gene_type,
                                                                                         sample_size=self.sample_size,
                                                                                         mutation_by_replacement=self.mutation_by_replacement,
                                                                                         build_initial_pop=False)
        return offspring

    def mutation_probs_by_space(self, offspring):

        """
        Applies the random mutation using the mutation values' space and the mutation probability. For each gene, if its probability is <= that the mutation probability, then it will be mutated based on the mutation space.
        It accepts:
            -offspring: The offspring to mutate.
        It returns an array of the mutated offspring using the mutation space.
        """

        # For each offspring, a value from the gene space is selected randomly and assigned to the selected mutated gene.
        for offspring_idx in range(offspring.shape[0]):
            probs = numpy.random.random(size=offspring.shape[1])
            for gene_idx in range(offspring.shape[1]):

                if probs[gene_idx] <= self.mutation_probability:
                    value_from_space = self.mutation_process_gene_value(solution=offspring[offspring_idx],
                                                                        gene_idx=gene_idx,
                                                                        sample_size=self.sample_size)

                    # Assigning the selected value from the space to the gene.
                    if self.gene_type_single == True:
                        if not self.gene_type[1] is None:
                            offspring[offspring_idx, gene_idx] = numpy.round(self.gene_type[0](value_from_space),
                                                                             self.gene_type[1])
                        else:
                            offspring[offspring_idx, gene_idx] = self.gene_type[0](value_from_space)
                    else:
                        if not self.gene_type[gene_idx][1] is None:
                            offspring[offspring_idx, gene_idx] = numpy.round(self.gene_type[gene_idx][0](value_from_space),
                                                                             self.gene_type[gene_idx][1])
                        else:
                            offspring[offspring_idx, gene_idx] = self.gene_type[gene_idx][0](value_from_space)

                    if self.allow_duplicate_genes == False:
                        offspring[offspring_idx], _, _ = self.solve_duplicate_genes_by_space(solution=offspring[offspring_idx],
                                                                                             gene_type=self.gene_type,
                                                                                             sample_size=self.sample_size,
                                                                                             mutation_by_replacement=self.mutation_by_replacement,
                                                                                             build_initial_pop=False)
        return offspring

    def mutation_process_gene_value(self,
                                    solution,
                                    gene_idx,
                                    range_min=None,
                                    range_max=None,
                                    sample_size=100):

        """
        Generate/select values for the gene that satisfy the constraint. The values could be generated randomly or from the gene space.
        It accepts:
            -range_min: The minimum value in the range from which a value is selected.
            -range_max: The maximum value in the range from which a value is selected.
            -solution: The solution where the target gene exists.
            -gene_idx: The index of the gene in the solution.
            -sample_size: The number of random values to generate from which a value is selected. It tries to generate a number of values up to a maximum of sample_size. But it is not always guaranteed because the total number of values might not be enough or the random generator creates duplicate random values.
        It returns a single numeric value the satisfies the gene constraint if exists in the gene_constraint parameter. 
        """

        # Check if the gene has a constraint.
        if self.gene_constraint and self.gene_constraint[gene_idx]:
            # Generate values that meet the gene constraint. Select more than 1 value.
            # This method: 1) generates or selects the values 2) filters the values according to the constraint.
            values = self.get_valid_gene_constraint_values(range_min=range_min,
                                                           range_max=range_max,
                                                           gene_value=solution[gene_idx],
                                                           gene_idx=gene_idx,
                                                           mutation_by_replacement=self.mutation_by_replacement,
                                                           solution=solution,
                                                           sample_size=sample_size)
            if values is None:
                # No value found that satisfy the constraint.
                # Keep the old value.
                value_selected = solution[gene_idx]
            else:
                # Select a value randomly from the list of values satisfying the constraint.
                # If size is used with numpy.random.choice(), it returns an array even if it has a single value. To return a numeric value, not an array, then return index 0.
                value_selected = numpy.random.choice(values, size=1)[0]
        else:
            # The gene does not have a constraint. Just select a single value.
            value_selected = self.generate_gene_value(range_min=range_min,
                                                      range_max=range_max,
                                                      gene_value=solution[gene_idx],
                                                      gene_idx=gene_idx,
                                                      solution=solution,
                                                      mutation_by_replacement=self.mutation_by_replacement,
                                                      sample_size=1)
        # Even that its name is singular, it might have a multiple values.
        return value_selected

    def mutation_randomly(self, offspring):

        """
        Applies the random mutation.
        It accepts:
            -offspring: The offspring to mutate.
        It returns an array of the mutated offspring.
        """

        # Random mutation changes one or more genes in each offspring randomly.
        for offspring_idx in range(offspring.shape[0]):
            # Return the indices of the genes to mutate.
            mutation_indices = numpy.array(random.sample(range(0, self.num_genes), 
                                                         self.mutation_num_genes))
            for gene_idx in mutation_indices:

                range_min, range_max = self.get_random_mutation_range(gene_idx)

                # Generate a random value for mutation that meet the gene constraint if exists.
                random_value = self.mutation_process_gene_value(range_min=range_min,
                                                                range_max=range_max,
                                                                solution=offspring[offspring_idx],
                                                                gene_idx=gene_idx,
                                                                sample_size=self.sample_size)

                offspring[offspring_idx, gene_idx] = random_value

                if self.allow_duplicate_genes == False:
                    offspring[offspring_idx], _, _ = self.solve_duplicate_genes_randomly(solution=offspring[offspring_idx],
                                                                                         min_val=range_min,
                                                                                         max_val=range_max,
                                                                                         mutation_by_replacement=self.mutation_by_replacement,
                                                                                         gene_type=self.gene_type,
                                                                                         sample_size=self.sample_size)

        return offspring

    def mutation_probs_randomly(self, offspring):

        """
        Applies the random mutation using the mutation probability. For each gene, if its probability is <= that mutation probability, then it will be mutated randomly.
        It accepts:
            -offspring: The offspring to mutate.
        It returns an array of the mutated offspring.
        """

        # Random mutation changes one or more genes in each offspring randomly.
        for offspring_idx in range(offspring.shape[0]):
            # The mutation probabilities for the current offspring.
            probs = numpy.random.random(size=offspring.shape[1])
            for gene_idx in range(offspring.shape[1]):

                range_min, range_max = self.get_random_mutation_range(gene_idx)

                # A gene is mutated only if its mutation probability is less than or equal to the threshold.
                if probs[gene_idx] <= self.mutation_probability:

                    # Generate a random value fpr mutation that meet the gene constraint if exists.
                    random_value = self.mutation_process_gene_value(range_min=range_min,
                                                                    range_max=range_max,
                                                                    solution=offspring[offspring_idx],
                                                                    gene_idx=gene_idx,
                                                                    sample_size=self.sample_size)

                    offspring[offspring_idx, gene_idx] = random_value

                    if self.allow_duplicate_genes == False:
                        offspring[offspring_idx], _, _ = self.solve_duplicate_genes_randomly(solution=offspring[offspring_idx],
                                                                                             min_val=range_min,
                                                                                             max_val=range_max,
                                                                                             mutation_by_replacement=self.mutation_by_replacement,
                                                                                             gene_type=self.gene_type,
                                                                                             sample_size=self.sample_size)
        return offspring

    def swap_mutation(self, offspring):

        """
        Applies the swap mutation which interchanges the values of 2 randomly selected genes.
        It accepts:
            -offspring: The offspring to mutate.
        It returns an array of the mutated offspring.
        """

        for idx in range(offspring.shape[0]):
            mutation_gene1 = numpy.random.randint(low=0, high=offspring.shape[1]/2, size=1)[0]
            mutation_gene2 = mutation_gene1 + int(offspring.shape[1]/2)

            temp = offspring[idx, mutation_gene1]
            offspring[idx, mutation_gene1] = offspring[idx, mutation_gene2]
            offspring[idx, mutation_gene2] = temp
        return offspring

    def inversion_mutation(self, offspring):

        """
        Applies the inversion mutation which selects a subset of genes and inverts them (in order).
        It accepts:
            -offspring: The offspring to mutate.
        It returns an array of the mutated offspring.
        """

        for idx in range(offspring.shape[0]):
            mutation_gene1 = numpy.random.randint(low=0, high=numpy.ceil(offspring.shape[1]/2 + 1), size=1)[0]
            mutation_gene2 = mutation_gene1 + int(offspring.shape[1]/2)

            genes_to_scramble = numpy.flip(offspring[idx, mutation_gene1:mutation_gene2])
            offspring[idx, mutation_gene1:mutation_gene2] = genes_to_scramble
        return offspring

    def scramble_mutation(self, offspring):

        """
        Applies the scramble mutation which selects a subset of genes and shuffles their order randomly.
        It accepts:
            -offspring: The offspring to mutate.
        It returns an array of the mutated offspring.
        """

        for idx in range(offspring.shape[0]):
            mutation_gene1 = numpy.random.randint(low=0, high=numpy.ceil(offspring.shape[1]/2 + 1), size=1)[0]
            mutation_gene2 = mutation_gene1 + int(offspring.shape[1]/2)
            genes_range = numpy.arange(start=mutation_gene1, stop=mutation_gene2)
            numpy.random.shuffle(genes_range)
            
            genes_to_scramble = numpy.flip(offspring[idx, genes_range])
            offspring[idx, genes_range] = genes_to_scramble
        return offspring

    def adaptive_mutation_population_fitness(self, offspring):

        """
        A helper method to calculate the average fitness of the solutions before applying the adaptive mutation.
        It accepts:
            -offspring: The offspring to mutate.
        It returns the average fitness to be used in adaptive mutation.
        """        

        fitness = self.last_generation_fitness.copy()
        temp_population = numpy.zeros_like(self.population)

        if (self.keep_elitism == 0):
            if (self.keep_parents == 0):
                parents_to_keep = []
            elif (self.keep_parents == -1):
                parents_to_keep = self.last_generation_parents.copy()
                temp_population[0:len(parents_to_keep), :] = parents_to_keep
            elif (self.keep_parents > 0):
                parents_to_keep, _ = self.steady_state_selection(self.last_generation_fitness, num_parents=self.keep_parents)
                temp_population[0:len(parents_to_keep), :] = parents_to_keep
        else:
            parents_to_keep, _ = self.steady_state_selection(self.last_generation_fitness, num_parents=self.keep_elitism)
            temp_population[0:len(parents_to_keep), :] = parents_to_keep

        temp_population[len(parents_to_keep):, :] = offspring

        fitness[:self.last_generation_parents.shape[0]] = self.last_generation_fitness[self.last_generation_parents_indices]

        first_idx = len(parents_to_keep)
        last_idx = fitness.shape[0]
        if len(fitness.shape) > 1:
            # TODO This is a multi-objective optimization problem.
            # fitness[first_idx:last_idx] = [0]*(last_idx - first_idx)
            fitness[first_idx:last_idx] = numpy.zeros(shape=(last_idx - first_idx, fitness.shape[1]))
            # raise ValueError('Edit adaptive mutation to work with multi-objective optimization problems.')
        else:
            # This is a single-objective optimization problem.
            fitness[first_idx:last_idx] = [0]*(last_idx - first_idx)

        # # No parallel processing.
        if self.parallel_processing is None:
            if self.fitness_batch_size in [1, None]:
                # Calculate the fitness for each individual solution.
                for idx in range(first_idx, last_idx):
                    # We cannot return the index of the solution within the population.
                    # Because the new solution (offspring) does not yet exist in the population.
                    # The user should handle this situation if the solution index is used anywhere.
                    fitness[idx] = self.fitness_func(self, 
                                                      temp_population[idx], 
                                                      None)
            else:
                # Calculate the fitness for batch of solutions.
    
                # Number of batches.
                num_batches = int(numpy.ceil((last_idx - first_idx) / self.fitness_batch_size))
    
                for batch_idx in range(num_batches):
                    # The index of the first solution in the current batch.
                    batch_first_index = first_idx + batch_idx * self.fitness_batch_size
                    # The index of the last solution in the current batch.
                    if batch_idx == (num_batches - 1):
                        batch_last_index = last_idx
                    else:
                        batch_last_index = first_idx + (batch_idx + 1) * self.fitness_batch_size
    
                    # Calculate the fitness values for the batch.
                    # We cannot return the index/indices of the solution(s) within the population.
                    # Because the new solution(s) (offspring) do(es) not yet exist in the population.
                    # The user should handle this situation if the solution index is used anywhere.
                    fitness_temp = self.fitness_func(self, 
                                                     temp_population[batch_first_index:batch_last_index], 
                                                     None) 
                    # Insert the fitness of each solution at the proper index.
                    for idx in range(batch_first_index, batch_last_index):
                        fitness[idx] = fitness_temp[idx - batch_first_index]

        else:
            # Parallel processing
            # Decide which class to use based on whether the user selected "process" or "thread"
            # TODO Add ExecutorClass as an instance attribute in the pygad.GA instances. Then retrieve this instance here instead of creating a new one.
            if self.parallel_processing[0] == "process":
                ExecutorClass = concurrent.futures.ProcessPoolExecutor
            else:
                ExecutorClass = concurrent.futures.ThreadPoolExecutor
    
            # We can use a with statement to ensure threads are cleaned up promptly (https://docs.python.org/3/library/concurrent.futures.html#threadpoolexecutor-example)
            with ExecutorClass(max_workers=self.parallel_processing[1]) as executor:
                # Indices of the solutions to calculate its fitness.
                solutions_to_submit_indices = list(range(first_idx, last_idx))
                # The solutions to calculate its fitness.
                solutions_to_submit = [temp_population[sol_idx].copy() for sol_idx in solutions_to_submit_indices]
                if self.fitness_batch_size in [1, None]:
                    # Use parallel processing to calculate the fitness of the solutions.
                    for index, sol_fitness in zip(solutions_to_submit_indices, executor.map(self.fitness_func, [self]*len(solutions_to_submit_indices), solutions_to_submit, solutions_to_submit_indices)):
                        if type(sol_fitness) in self.supported_int_float_types:
                            # The fitness function returns a single numeric value.
                            # This is a single-objective optimization problem.
                            fitness[index] = sol_fitness
                        elif type(sol_fitness) in [list, tuple, numpy.ndarray]:
                            # The fitness function returns a list/tuple/numpy.ndarray.
                            # This is a multi-objective optimization problem.
                            fitness[index] = sol_fitness
                        else:
                            raise ValueError(f"The fitness function should return a number or an iterable (list, tuple, or numpy.ndarray) but the value {sol_fitness} of type {type(sol_fitness)} found.")
                else:
                    # Reaching this point means that batch processing is in effect to calculate the fitness values.
                    # Number of batches.
                    num_batches = int(numpy.ceil(len(solutions_to_submit_indices) / self.fitness_batch_size))
                    # Each element of the `batches_solutions` list represents the solutions in one batch.
                    batches_solutions = []
                    # Each element of the `batches_indices` list represents the solutions' indices in one batch.
                    batches_indices = []
                    # For each batch, get its indices and call the fitness function.
                    for batch_idx in range(num_batches):
                        batch_first_index = batch_idx * self.fitness_batch_size
                        batch_last_index = (batch_idx + 1) * self.fitness_batch_size
                        batch_indices = solutions_to_submit_indices[batch_first_index:batch_last_index]
                        batch_solutions = self.population[batch_indices, :]
    
                        batches_solutions.append(batch_solutions)
                        batches_indices.append(batch_indices)

                    for batch_indices, batch_fitness in zip(batches_indices, executor.map(self.fitness_func, [self]*len(solutions_to_submit_indices), batches_solutions, batches_indices)):
                        if type(batch_fitness) not in [list, tuple, numpy.ndarray]:
                            raise TypeError(f"Expected to receive a list, tuple, or numpy.ndarray from the fitness function but the value ({batch_fitness}) of type {type(batch_fitness)}.")
                        elif len(numpy.array(batch_fitness)) != len(batch_indices):
                            raise ValueError(f"There is a mismatch between the number of solutions passed to the fitness function ({len(batch_indices)}) and the number of fitness values returned ({len(batch_fitness)}). They must match.")

                        for index, sol_fitness in zip(batch_indices, batch_fitness):
                            if type(sol_fitness) in self.supported_int_float_types:
                                # The fitness function returns a single numeric value.
                                # This is a single-objective optimization problem.
                                fitness[index] = sol_fitness
                            elif type(sol_fitness) in [list, tuple, numpy.ndarray]:
                                # The fitness function returns a list/tuple/numpy.ndarray.
                                # This is a multi-objective optimization problem.
                                fitness[index] = sol_fitness
                            else:
                                raise ValueError(f"The fitness function should return a number or an iterable (list, tuple, or numpy.ndarray) but the value ({sol_fitness}) of type {type(sol_fitness)} found.")

        if len(fitness.shape) > 1:
            # TODO This is a multi-objective optimization problem.
            # Calculate the average of each objective's fitness across all solutions in the population.
            average_fitness = numpy.mean(fitness, axis=0)
        else:
            # This is a single-objective optimization problem.
            average_fitness = numpy.mean(fitness)

        return average_fitness, fitness[len(parents_to_keep):]

    def adaptive_mutation(self, offspring):

        """
        Applies the adaptive mutation which changes the values of a number of genes randomly. In adaptive mutation, the number of genes to mutate differs based on the fitness value of the solution.
        The random value is selected either using the 'gene_space' parameter or the 2 parameters 'random_mutation_min_val' and 'random_mutation_max_val'.
        It accepts:
            -offspring: The offspring to mutate.
        It returns an array of the mutated offspring.
        """

        # If the attribute 'gene_space' exists (i.e. not None), then the mutation values are selected from the 'gene_space' parameter according to the space of values of each gene. Otherwise, it is selected randomly based on the 2 parameters 'random_mutation_min_val' and 'random_mutation_max_val'.
        # When the 'mutation_probability' parameter exists (i.e. not None), then it is used in the mutation. Otherwise, the 'mutation_num_genes' parameter is used.

        if self.mutation_probability is None:
            # When the 'mutation_probability' parameter does not exist (i.e. None), then the parameter 'mutation_num_genes' is used in the mutation.
            if not (self.gene_space is None):
                # When the attribute 'gene_space' exists (i.e. not None), the mutation values are selected randomly from the space of values of each gene.
                offspring = self.adaptive_mutation_by_space(offspring)
            else:
                # When the attribute 'gene_space' does not exist (i.e. None), the mutation values are selected randomly based on the continuous range specified by the 2 attributes 'random_mutation_min_val' and 'random_mutation_max_val'.
                offspring = self.adaptive_mutation_randomly(offspring)
        else:
            # When the 'mutation_probability' parameter exists (i.e. not None), then it is used in the mutation.
            if not (self.gene_space is None):
                # When the attribute 'gene_space' exists (i.e. not None), the mutation values are selected randomly from the space of values of each gene.
                offspring = self.adaptive_mutation_probs_by_space(offspring)
            else:
                # When the attribute 'gene_space' does not exist (i.e. None), the mutation values are selected randomly based on the continuous range specified by the 2 attributes 'random_mutation_min_val' and 'random_mutation_max_val'.
                offspring = self.adaptive_mutation_probs_randomly(offspring)

        return offspring

    def adaptive_mutation_by_space(self, offspring):

        """
        Applies the adaptive mutation based on the 2 parameters 'mutation_num_genes' and 'gene_space'. 
        A number of genes equal are selected randomly for mutation. This number depends on the fitness of the solution.
        The random values are selected from the 'gene_space' parameter.
        It accepts:
            -offspring: The offspring to mutate.
        It returns an array of the mutated offspring.
        """
        
        # For each offspring, a value from the gene space is selected randomly and assigned to the selected gene for mutation.

        average_fitness, offspring_fitness = self.adaptive_mutation_population_fitness(offspring)

        # Adaptive mutation changes one or more genes in each offspring randomly.
        # The number of genes to mutate depends on the solution's fitness value.
        for offspring_idx in range(offspring.shape[0]):
            ## TODO Make edits to work with multi-objective optimization.
            # Compare the fitness of each offspring to the average fitness of each objective function.
            fitness_comparison = offspring_fitness[offspring_idx] < average_fitness

            # Check if the problem is single or multi-objective optimization.
            if type(fitness_comparison) in [bool, numpy.bool_]:
                # Single-objective optimization problem.
                if offspring_fitness[offspring_idx] < average_fitness:
                    adaptive_mutation_num_genes = self.mutation_num_genes[0]
                else:
                    adaptive_mutation_num_genes = self.mutation_num_genes[1]
            else:
                # Multi-objective optimization problem.

                # Get the sum of the pool array (result of comparison).
                # True is considered 1 and False is 0.
                fitness_comparison_sum = sum(fitness_comparison)
                # Check if more than or equal to 50% of the objectives have fitness greater than the average.
                # If True, then use the first percentage. 
                # If False, use the second percentage.
                if fitness_comparison_sum >= len(fitness_comparison)/2:
                    adaptive_mutation_num_genes = self.mutation_num_genes[0]
                else:
                    adaptive_mutation_num_genes = self.mutation_num_genes[1]

            mutation_indices = numpy.array(random.sample(range(0, self.num_genes), adaptive_mutation_num_genes))
            for gene_idx in mutation_indices:

                value_from_space = self.mutation_process_gene_value(solution=offspring[offspring_idx],
                                                                    gene_idx=gene_idx,
                                                                    sample_size=self.sample_size)

                # Assigning the selected value from the space to the gene.
                if self.gene_type_single == True:
                    if not self.gene_type[1] is None:
                        offspring[offspring_idx, gene_idx] = numpy.round(self.gene_type[0](value_from_space),
                                                                         self.gene_type[1])
                    else:
                        offspring[offspring_idx, gene_idx] = self.gene_type[0](value_from_space)
                else:
                    if not self.gene_type[gene_idx][1] is None:
                        offspring[offspring_idx, gene_idx] = numpy.round(self.gene_type[gene_idx][0](value_from_space),
                                                                         self.gene_type[gene_idx][1])
                    else:
                        offspring[offspring_idx, gene_idx] = self.gene_type[gene_idx][0](value_from_space)

                if self.allow_duplicate_genes == False:
                    offspring[offspring_idx], _, _ = self.solve_duplicate_genes_by_space(solution=offspring[offspring_idx],
                                                                                         gene_type=self.gene_type,
                                                                                         sample_size=self.sample_size,
                                                                                         mutation_by_replacement=self.mutation_by_replacement,
                                                                                         build_initial_pop=False)
        return offspring

    def adaptive_mutation_randomly(self, offspring):

        """
        Applies the adaptive mutation based on the 'mutation_num_genes' parameter. 
        A number of genes equal are selected randomly for mutation. This number depends on the fitness of the solution.
        The random values are selected based on the 2 parameters 'random_mutation_min_val' and 'random_mutation_max_val'.
        It accepts:
            -offspring: The offspring to mutate.
        It returns an array of the mutated offspring.
        """

        average_fitness, offspring_fitness = self.adaptive_mutation_population_fitness(offspring)

        # Adaptive random mutation changes one or more genes in each offspring randomly.
        # The number of genes to mutate depends on the solution's fitness value.
        for offspring_idx in range(offspring.shape[0]):
            ## TODO Make edits to work with multi-objective optimization.
            # Compare the fitness of each offspring to the average fitness of each objective function.
            fitness_comparison = offspring_fitness[offspring_idx] < average_fitness

            # Check if the problem is single or multi-objective optimization.
            if type(fitness_comparison) in [bool, numpy.bool_]:
                # Single-objective optimization problem.
                if fitness_comparison:
                    adaptive_mutation_num_genes = self.mutation_num_genes[0]
                else:
                    adaptive_mutation_num_genes = self.mutation_num_genes[1]
            else:
                # Multi-objective optimization problem.

                # Get the sum of the pool array (result of comparison).
                # True is considered 1 and False is 0.
                fitness_comparison_sum = sum(fitness_comparison)
                # Check if more than or equal to 50% of the objectives have fitness greater than the average.
                # If True, then use the first percentage. 
                # If False, use the second percentage.
                if fitness_comparison_sum >= len(fitness_comparison)/2:
                    adaptive_mutation_num_genes = self.mutation_num_genes[0]
                else:
                    adaptive_mutation_num_genes = self.mutation_num_genes[1]

            mutation_indices = numpy.array(random.sample(range(0, self.num_genes), adaptive_mutation_num_genes))
            for gene_idx in mutation_indices:

                range_min, range_max = self.get_random_mutation_range(gene_idx)

                # Generate a random value fpr mutation that meet the gene constraint if exists.
                random_value = self.mutation_process_gene_value(range_min=range_min,
                                                                range_max=range_max,
                                                                solution=offspring[offspring_idx],
                                                                gene_idx=gene_idx,
                                                                sample_size=self.sample_size)

                offspring[offspring_idx, gene_idx] = random_value

                if self.allow_duplicate_genes == False:
                    offspring[offspring_idx], _, _ = self.solve_duplicate_genes_randomly(solution=offspring[offspring_idx],
                                                                                         min_val=range_min,
                                                                                         max_val=range_max,
                                                                                         mutation_by_replacement=self.mutation_by_replacement,
                                                                                         gene_type=self.gene_type,
                                                                                         sample_size=self.sample_size)
        return offspring

    def adaptive_mutation_probs_by_space(self, offspring):

        """
        Applies the adaptive mutation based on the 2 parameters 'mutation_probability' and 'gene_space'.
        Based on whether the solution fitness is above or below a threshold, the mutation is applied diffrently by mutating high or low number of genes.
        The random values are selected based on space of values for each gene.
        It accepts:
            -offspring: The offspring to mutate.
        It returns an array of the mutated offspring.
        """

        # For each offspring, a value from the gene space is selected randomly and assigned to the selected gene for mutation.

        average_fitness, offspring_fitness = self.adaptive_mutation_population_fitness(offspring)

        # Adaptive random mutation changes one or more genes in each offspring randomly.
        # The probability of mutating a gene depends on the solution's fitness value.
        for offspring_idx in range(offspring.shape[0]):
            ## TODO Make edits to work with multi-objective optimization.
            # Compare the fitness of each offspring to the average fitness of each objective function.
            fitness_comparison = offspring_fitness[offspring_idx] < average_fitness

            # Check if the problem is single or multi-objective optimization.
            if type(fitness_comparison) in [bool, numpy.bool_]:
                # Single-objective optimization problem.
                if offspring_fitness[offspring_idx] < average_fitness:
                    adaptive_mutation_probability = self.mutation_probability[0]
                else:
                    adaptive_mutation_probability = self.mutation_probability[1]
            else:
                # Multi-objective optimization problem.

                # Get the sum of the pool array (result of comparison).
                # True is considered 1 and False is 0.
                fitness_comparison_sum = sum(fitness_comparison)
                # Check if more than or equal to 50% of the objectives have fitness greater than the average.
                # If True, then use the first percentage. 
                # If False, use the second percentage.
                if fitness_comparison_sum >= len(fitness_comparison)/2:
                    adaptive_mutation_probability = self.mutation_probability[0]
                else:
                    adaptive_mutation_probability = self.mutation_probability[1]

            probs = numpy.random.random(size=offspring.shape[1])
            for gene_idx in range(offspring.shape[1]):

                if probs[gene_idx] <= adaptive_mutation_probability:

                    value_from_space = self.mutation_process_gene_value(solution=offspring[offspring_idx],
                                                                        gene_idx=gene_idx,
                                                                        sample_size=self.sample_size)

                    # Assigning the selected value from the space to the gene.
                    if self.gene_type_single == True:
                        if not self.gene_type[1] is None:
                            offspring[offspring_idx, gene_idx] = numpy.round(self.gene_type[0](value_from_space),
                                                                             self.gene_type[1])
                        else:
                            offspring[offspring_idx, gene_idx] = self.gene_type[0](value_from_space)
                    else:
                        if not self.gene_type[gene_idx][1] is None:
                            offspring[offspring_idx, gene_idx] = numpy.round(self.gene_type[gene_idx][0](value_from_space),
                                                                             self.gene_type[gene_idx][1])
                        else:
                            offspring[offspring_idx, gene_idx] = self.gene_type[gene_idx][0](value_from_space)

                    if self.allow_duplicate_genes == False:
                        offspring[offspring_idx], _, _ = self.solve_duplicate_genes_by_space(solution=offspring[offspring_idx],
                                                                                             gene_type=self.gene_type,
                                                                                             sample_size=self.sample_size,
                                                                                             mutation_by_replacement=self.mutation_by_replacement,
                                                                                             build_initial_pop=False)
        return offspring

    def adaptive_mutation_probs_randomly(self, offspring):

        """
        Applies the adaptive mutation based on the 'mutation_probability' parameter. 
        Based on whether the solution fitness is above or below a threshold, the mutation is applied diffrently by mutating high or low number of genes.
        The random values are selected based on the 2 parameters 'random_mutation_min_val' and 'random_mutation_max_val'.
        It accepts:
            -offspring: The offspring to mutate.
        It returns an array of the mutated offspring.
        """

        average_fitness, offspring_fitness = self.adaptive_mutation_population_fitness(offspring)

        # Adaptive random mutation changes one or more genes in each offspring randomly.
        # The probability of mutating a gene depends on the solution's fitness value.
        for offspring_idx in range(offspring.shape[0]):
            ## TODO Make edits to work with multi-objective optimization.
            # Compare the fitness of each offspring to the average fitness of each objective function.
            fitness_comparison = offspring_fitness[offspring_idx] < average_fitness

            # Check if the problem is single or multi-objective optimization.
            if type(fitness_comparison) in [bool, numpy.bool_]:
                # Single-objective optimization problem.
                if offspring_fitness[offspring_idx] < average_fitness:
                    adaptive_mutation_probability = self.mutation_probability[0]
                else:
                    adaptive_mutation_probability = self.mutation_probability[1]
            else:
                # Multi-objective optimization problem.

                # Get the sum of the pool array (result of comparison).
                # True is considered 1 and False is 0.
                fitness_comparison_sum = sum(fitness_comparison)
                # Check if more than or equal to 50% of the objectives have fitness greater than the average.
                # If True, then use the first percentage. 
                # If False, use the second percentage.
                if fitness_comparison_sum >= len(fitness_comparison)/2:
                    adaptive_mutation_probability = self.mutation_probability[0]
                else:
                    adaptive_mutation_probability = self.mutation_probability[1]

            probs = numpy.random.random(size=offspring.shape[1])
            for gene_idx in range(offspring.shape[1]):

                range_min, range_max = self.get_random_mutation_range(gene_idx)

                if probs[gene_idx] <= adaptive_mutation_probability:
                    # Generate a random value fpr mutation that meet the gene constraint if exists.
                    random_value = self.mutation_process_gene_value(range_min=range_min,
                                                                    range_max=range_max,
                                                                    solution=offspring[offspring_idx],
                                                                    gene_idx=gene_idx,
                                                                    sample_size=self.sample_size)

                    offspring[offspring_idx, gene_idx] = random_value

                    if self.allow_duplicate_genes == False:
                        offspring[offspring_idx], _, _ = self.solve_duplicate_genes_randomly(solution=offspring[offspring_idx],
                                                                                             min_val=range_min,
                                                                                             max_val=range_max,
                                                                                             mutation_by_replacement=self.mutation_by_replacement,
                                                                                             gene_type=self.gene_type,
                                                                                             sample_size=self.sample_size)
        return offspring



"""
The pygad.helper.misc module has some generic helper methods.
"""

import numpy
import warnings
import random
import pygad

class Helper:

    def change_population_dtype_and_round(self,
                                          population):
        """
        Change the data type of the population. It works with iterables (e.g. lists or NumPy arrays) of shape 2D.
        It does not handle single numeric values or 1D arrays.

        It accepts:
            -population: The iterable to change its dtype.

        It returns the iterable with the data type changed for all genes.
        """

        population_new = numpy.array(population.copy(), dtype=object)

        # Forcing the iterable to have the data type assigned to the gene_type parameter.
        if self.gene_type_single == True:
            # Round the numbers first then change the data type.
            # This solves issues with some data types such as numpy.float32.
            if self.gene_type[1] is None:
                pass
            else:
                # This block is reached only for non-integer data types (i.e. float).
                population_new = numpy.round(numpy.array(population_new, float),
                                             self.gene_type[1])

            population_new = numpy.array(population_new,
                                         dtype=self.gene_type[0])
        else:
            population = numpy.array(population.copy())
            population_new = numpy.zeros(shape=population.shape,
                                         dtype=object)
            for gene_idx in range(population.shape[1]):
                # Round the numbers first then change the data type.
                # This solves issues with some data types such as numpy.float32.
                if self.gene_type[gene_idx][1] is None:
                    # Do not round.
                    population_new[:, gene_idx] = population[:, gene_idx]
                else:
                    # This block is reached only for non-integer data types (i.e. float).
                    population_new[:, gene_idx] = numpy.round(numpy.array(population[:, gene_idx], float),
                                                              self.gene_type[gene_idx][1])
                # Once rounding is done, change the data type.
                # population_new[:, gene_idx] = numpy.asarray(population_new[:, gene_idx], dtype=self.gene_type[gene_idx][0])
                # Use a for loop to maintain the data type of each individual gene.
                for sol_idx in range(population.shape[0]):
                    population_new[sol_idx, gene_idx] = self.gene_type[gene_idx][0](population_new[sol_idx, gene_idx])
        return population_new

    def change_gene_dtype_and_round(self,
                                    gene_index,
                                    gene_value):
        """
        Change the data type and round a single gene value or a vector of values FOR THE SAME GENE. E.g., the input could be 6 or [6, 7, 8].

        It accepts 2 parameters:
            -gene_index: The index of the target gene.
            -gene_value: The gene value.

        If gene_value has a single value, then it returns a single number with the type changed and value rounded. If gene_value is a vector, then a vector is returned after changing the data type and rounding.
        """

        if self.gene_type_single == True:
            dtype = self.gene_type[0]
            if self.gene_type[1] is None:
                # No rounding for this gene. Use the old gene value.
                round_precision = None
            else:
                round_precision = self.gene_type[1]
        else:
            dtype = self.gene_type[gene_index][0]
            if self.gene_type[gene_index][1] is None:
                # No rounding for this gene. Use the old gene value.
                round_precision = None
            else:
                round_precision = self.gene_type[gene_index][1]

        # Sometimes the values represent the gene_space when it is not nested (e.g. gene_space=range(10))
        # Copy it to avoid changing the original gene_space.
        gene_value = [gene_value].copy()

        # Round the number before changing its data type to avoid precision loss for some data types like numpy.float32.
        if round_precision is None:
            pass
        else:
            gene_value = numpy.round(gene_value, round_precision)

        gene_value_new = numpy.asarray(gene_value, dtype=dtype)
        gene_value_new = gene_value_new[0]

        return gene_value_new

    def mutation_change_gene_dtype_and_round(self,
                                             random_value,
                                             gene_index,
                                             gene_value,
                                             mutation_by_replacement):
        """
        Change the data type and round the random value used to apply mutation.

        It accepts:
            -random_value: The random value to change its data type.
            -gene_index: The index of the target gene.
            -gene_value: The gene value before mutation. Only used if mutation_by_replacement=False and gene_type_single=False.
            -mutation_by_replacement: A flag indicating whether mutation by replacement is enabled or not. The reason is to make this helper method usable while generating the initial population. In this case, mutation_by_replacement does not matter and should be considered False.

        It returns the new value after changing the data type and being rounded.
        """

        if mutation_by_replacement:
            # If the mutation_by_replacement attribute is True, then the random value replaces the current gene value.
            gene_value = random_value
        else:
            # If the mutation_by_replacement attribute is False, then the random value is added to the gene value.
            gene_value = gene_value + random_value

        gene_value_new = self.change_gene_dtype_and_round(gene_index=gene_index,
                                                          gene_value=gene_value)
        return gene_value_new

    def validate_gene_constraint_callable_output(self,
                                                 selected_values,
                                                 values):
        if type(selected_values) in [list, numpy.ndarray]:
            selected_values_set = set(selected_values)
            if selected_values_set.issubset(values):
                pass
            else:
                return False
        else:
            return False

        return True

    def filter_gene_values_by_constraint(self,
                                         values,
                                         solution,
                                         gene_idx):

        """
        Filter the random values generated for mutation based on whether they meet the gene constraint in the gene_constraint parameter.

        It accepts:
            -values: The values to filter.
            -solution: The solution containing the target gene.
            -gene_idx: The index of the gene in the solution.

        It returns None if no values satisfy the constraint. Otherwise, an array of values that satisfy the constraint is returned.
        """

        if self.gene_constraint and self.gene_constraint[gene_idx]:
            pass
        else:
            raise Exception(f"Either the gene at index {gene_idx} is not assigned a callable/function or the gene_constraint itself is not used.")

        # A temporary solution to avoid changing the original solution.
        solution_tmp = solution.copy()
        filtered_values = self.gene_constraint[gene_idx](solution_tmp, values.copy())
        result = self.validate_gene_constraint_callable_output(selected_values=filtered_values,
                                                               values=values)
        if result:
            pass
        else:
            raise Exception("The output from the gene_constraint callable/function must be a list or NumPy array that is subset of the passed values (second argument).")

        # After going through all the values, check if any value satisfies the constraint.
        if len(filtered_values) > 0:
            # At least one value was found that meets the gene constraint.
            pass
        else:
            # No value found for the current gene that satisfies the constraint.
            if not self.suppress_warnings: warnings.warn(f"Failed to find a value that satisfies its gene constraint for the gene at index {gene_idx} with value {solution[gene_idx]} at generation {self.generations_completed}.")
            return None

        return filtered_values

    def get_gene_dtype(self, gene_index):

        """
        Returns the data type of the gene by its index.

        It accepts a single parameter:
            -gene_index: The index of the gene to get its data type. Only used if each gene has its own data type.

        It returns the data type of the gene.
        """

        if self.gene_type_single == True:
            dtype = self.gene_type
        else:
            dtype = self.gene_type[gene_index]
        return dtype

    def get_random_mutation_range(self, gene_index):

        """
        Returns the minimum and maximum values of the mutation range.

        It accepts a single parameter:
            -gene_index: The index of the gene to get its range. Only used if the gene has a specific mutation range.

        It returns the minimum and maximum values of the gene mutation range.
        """

        # We can use either random_mutation_min_val or random_mutation_max_val.
        if type(self.random_mutation_min_val) in self.supported_int_float_types:
            range_min = self.random_mutation_min_val
            range_max = self.random_mutation_max_val
        else:
            range_min = self.random_mutation_min_val[gene_index]
            range_max = self.random_mutation_max_val[gene_index]
        return range_min, range_max

    def get_initial_population_range(self, gene_index):

        """
        Returns the minimum and maximum values of the initial population range.

        It accepts a single parameter:
            -gene_index: The index of the gene to get its range. Only used if the gene has a specific range

        It returns the minimum and maximum values of the gene initial population range.
        """

        # We can use either init_range_low or init_range_high.
        if type(self.init_range_low) in self.supported_int_float_types:
            range_min = self.init_range_low
            range_max = self.init_range_high
        else:
            range_min = self.init_range_low[gene_index]
            range_max = self.init_range_high[gene_index]
        return range_min, range_max

    def generate_gene_value_from_space(self,
                                       gene_idx,
                                       mutation_by_replacement,
                                       solution=None,
                                       gene_value=None,
                                       sample_size=1):
        """
        Generate/select one or more values for the gene from the gene space.

        It accepts:
            -gene_idx: The index of the gene in the solution.
            -mutation_by_replacement: A flag indicating whether mutation by replacement is enabled or not. The reason is to make this helper method usable while generating the initial population. In this case, mutation_by_replacement does not matter and should be considered False.
            -solution (iterable, optional): The solution where we need to generate a gene. Needed if you are selecting a single value (sample_size=1) to select a value that respects the allow_duplicate_genes parameter instead of selecting a value randomly. If None, then the gene value is selected randomly.
            -gene_value (int, optional): The original gene value before applying mutation. Needed if you are calling this method to apply mutation. If None, then a sample is created from the gene space without being summed to the gene value.
            -sample_size (int, optional): The number of random values to generate. It tries to generate a number of values up to a maximum of sample_size. But it is not always guaranteed because the total number of values might not be enough or the random generator creates duplicate random values. For int data types, it could be None to keep all the values. For float data types, a None value returns only a single value.

        It returns,
            -A single numeric value if sample_size=1. Or
            -An array with number of maximum number of values equal to sample_size if sample_size>1.
        """

        if gene_value is None:
            # Use the initial population range.
            range_min, range_max = self.get_initial_population_range(gene_index=gene_idx)
        else:
            # Use the mutation range. 
            range_min, range_max = self.get_random_mutation_range(gene_idx)

        if self.gene_space_nested:
            # Returning the current gene space from the 'gene_space' attribute.
            # It is used to determine the way of selecting the next gene value:
            # 1) List/NumPy Array: Whether it has only one value, multiple values, or one of its values is None.
            # 2) Fixed Numeric Value
            # 3) None
            # 4) Dict: Whether the dict has the key `step` or not.
            if type(self.gene_space[gene_idx]) in [numpy.ndarray, list]:
                # Get the gene space from the `gene_space_unpacked` property because it undergoes data type change and rounded.
                curr_gene_space = self.gene_space_unpacked[gene_idx].copy()
            elif type(self.gene_space[gene_idx]) in pygad.GA.supported_int_float_types:
                # Get the gene space from the `gene_space_unpacked` property because it undergoes data type change and rounded.
                curr_gene_space = self.gene_space_unpacked[gene_idx]
            else:
                curr_gene_space = self.gene_space[gene_idx]

            if type(curr_gene_space) in pygad.GA.supported_int_float_types:
                # If the gene space is simply a single numeric value (e.g. 5), use it as the new gene value.
                value_from_space = curr_gene_space
            elif curr_gene_space is None:
                # If the gene space is None, apply mutation by adding a random value between the range defined by the 2 parameters 'random_mutation_min_val' and 'random_mutation_max_val'.
                rand_val = numpy.random.uniform(low=range_min,
                                                high=range_max,
                                                size=sample_size)
                if mutation_by_replacement:
                    value_from_space = rand_val
                else:
                    value_from_space = gene_value + rand_val
            elif type(curr_gene_space) is dict:
                # Selecting a value randomly from the current gene's space in the 'gene_space' attribute.
                # The gene's space of type dict specifies the lower and upper limits of a gene.
                if 'step' in curr_gene_space.keys():
                    # When the `size` parameter is used, the numpy.random.choice() and numpy.random.uniform() functions return a NumPy array as the output even if the array has a single value (i.e. size=1).
                    # We have to return the output at index 0 to force a numeric value to be returned not an object of type numpy.ndarray.
                    # If numpy.ndarray is returned, then it will cause an issue later while using the set() function.
                    # Randomly select a value from a discrete range.
                    value_from_space = numpy.random.choice(numpy.arange(start=curr_gene_space['low'],
                                                                        stop=curr_gene_space['high'],
                                                                        step=curr_gene_space['step']),
                                                           size=sample_size)
                else:
                    value_from_space = numpy.random.uniform(low=curr_gene_space['low'],
                                                            high=curr_gene_space['high'],
                                                            size=sample_size)
            else:
                # Selecting a value randomly from the current gene's space in the 'gene_space' attribute.
                # If the gene space has only 1 value, then select it. The old and new values of the gene are identical.
                if len(curr_gene_space) == 1:
                    value_from_space = curr_gene_space
                else:
                    # Change the data type and round the generated values.
                    curr_gene_space = self.change_gene_dtype_and_round(gene_index=gene_idx,
                                                                       gene_value=curr_gene_space)

                    if gene_value is None:
                        # Just generate the value(s) without being added to the gene value specially when initializing the population.
                        value_from_space = curr_gene_space
                    else:
                        # If the gene space has more than 1 value, then select a new one that is different from the current value.
                        # To avoid selecting the current gene value again, remove it from the current gene space and do the selection.
                        value_from_space = list(set(curr_gene_space) - set([gene_value]))
        else:
            # Selecting a value randomly from the global gene space in the 'gene_space' attribute.
            # The gene's space of type dict specifies the lower and upper limits of a gene.
            if type(self.gene_space) is dict:
                # When the gene_space is assigned a dict object, then it specifies the lower and upper limits of all genes in the space.
                if 'step' in self.gene_space.keys():
                    value_from_space = numpy.random.choice(numpy.arange(start=self.gene_space['low'],
                                                                        stop=self.gene_space['high'],
                                                                        step=self.gene_space['step']),
                                                           size=sample_size)
                else:
                    value_from_space = numpy.random.uniform(low=self.gene_space['low'],
                                                            high=self.gene_space['high'],
                                                            size=sample_size)
            else:
                curr_gene_space = list(self.gene_space).copy()
                for idx in range(len(curr_gene_space)):
                    if curr_gene_space[idx] is None:
                        curr_gene_space[idx] = numpy.random.uniform(low=range_min,
                                                                    high=range_max)
                curr_gene_space = self.change_gene_dtype_and_round(gene_index=gene_idx,
                                                                   gene_value=curr_gene_space)

                if gene_value is None:
                    # Just generate the value(s) without being added to the gene value specially when initializing the population.
                    value_from_space = curr_gene_space
                else:
                    # If the space type is not of type dict, then a value is randomly selected from the gene_space attribute.
                    # To avoid selecting the current gene value again, remove it from the current gene space and do the selection.
                    value_from_space = list(set(curr_gene_space) - set([gene_value]))

        if len(value_from_space) == 0:
            if gene_value is None:
                raise ValueError(f"There are no values to select from the gene_space for the gene at index {gene_idx}.")
            else:
                # After removing the current gene value from the space, there are no more values.
                # Then keep the current gene value.
                value_from_space = gene_value
                if sample_size > 1:
                    value_from_space = numpy.array([gene_value])
        elif sample_size == 1:
            if self.allow_duplicate_genes == True:
                # Select a value randomly from the current gene space.
                value_from_space = random.choice(value_from_space)
            else:
                # We must check if the selected value will respect the allow_duplicate_genes parameter.
                # Instead of selecting a value randomly, we have to select a value that will be unique if allow_duplicate_genes=False.
                # Only select a value from the current gene space that is, hopefully, unique.
                value_from_space = self.select_unique_value(gene_values=value_from_space,
                                                            solution=solution,
                                                            gene_index=gene_idx)

            # The gene space might be [None, 1, 7].
            # It might happen that the value None is selected.
            # In this case, generate a random value out of the mutation range.
            if value_from_space is None:
                value_from_space = numpy.random.uniform(low=range_min,
                                                        high=range_max,
                                                        size=sample_size)
        else:
            value_from_space = numpy.array(value_from_space)

        # Change the data type and round the generated values.
        # It has to be called here for all the missed cases.
        value_from_space = self.change_gene_dtype_and_round(gene_index=gene_idx,
                                                            gene_value=value_from_space)
        if sample_size == 1 and type(value_from_space) not in pygad.GA.supported_int_float_types:
            value_from_space = value_from_space[0]

        return value_from_space

    def generate_gene_value_randomly(self,
                                     range_min,
                                     range_max,
                                     gene_value,
                                     gene_idx,
                                     mutation_by_replacement,
                                     sample_size=1,
                                     step=1):
        """
        Randomly generate one or more values for the gene.
        It accepts:
            -range_min: The minimum value in the range from which a value is selected.
            -range_max: The maximum value in the range from which a value is selected.
            -gene_value: The original gene value before applying mutation.
            -gene_idx: The index of the gene in the solution.
            -mutation_by_replacement: A flag indicating whether mutation by replacement is enabled or not. The reason is to make this helper method usable while generating the initial population. In this case, mutation_by_replacement does not matter and should be considered False.
            -sample_size: The number of random values to generate. It tries to generate a number of values up to a maximum of sample_size. But it is not always guaranteed because the total number of values might not be enough or the random generator creates duplicate random values. For int data types, it could be None to keep all the values. For float data types, a None value returns only a single value.
            -step (int, optional): The step size for generating candidate values. Defaults to 1. Only used with genes of an integer data type.

        It returns,
            -A single numeric value if sample_size=1. Or
            -An array with number of values equal to sample_size if sample_size>1.
        """

        gene_type = self.get_gene_dtype(gene_index=gene_idx)
        if gene_type[0] in pygad.GA.supported_int_types:
            random_value = numpy.asarray(numpy.arange(range_min, 
                                                      range_max, 
                                                      step=step), 
                                         dtype=gene_type[0])
            if sample_size is None:
                # Keep all the values.
                pass
            else:
                if sample_size >= len(random_value):
                    # Number of values is larger than or equal to the number of elements in random_value.
                    # Makes no sense to create a larger sample out of the population because it just creates redundant values.
                    pass
                else:
                    # Set replace=True to avoid selecting the same value more than once.
                    random_value = numpy.random.choice(random_value, 
                                                       size=sample_size,
                                                       replace=False)
        else:
            # Generating a random value.
            random_value = numpy.asarray(numpy.random.uniform(low=range_min, 
                                                              high=range_max, 
                                                              size=sample_size),
                                         dtype=object)

        # Change the random mutation value data type.
        for idx, val in enumerate(random_value):
            random_value[idx] = self.mutation_change_gene_dtype_and_round(random_value[idx],
                                                                          gene_idx,
                                                                          gene_value,
                                                                          mutation_by_replacement=mutation_by_replacement)

        # Rounding different values could return the same value multiple times.
        # For example, 2.8 and 2.7 will be 3.0.
        # Use the unique() function to avoid any duplicates.
        random_value = numpy.unique(random_value)

        if sample_size == 1:
            random_value = random_value[0]

        return random_value

    def generate_gene_value(self,
                            gene_value,
                            gene_idx,
                            mutation_by_replacement,
                            solution=None,
                            range_min=None,
                            range_max=None,
                            sample_size=1,
                            step=1):
        """
        Generate one or more values for the gene either randomly or from the gene space. It acts as a router.
        It accepts:
            -gene_value: The original gene value before applying mutation.
            -gene_idx: The index of the gene in the solution.
            -mutation_by_replacement: A flag indicating whether mutation by replacement is enabled or not. The reason is to make this helper method usable while generating the initial population. In this case, mutation_by_replacement does not matter and should be considered False.
            -solution (iterable, optional): The solution where we need to generate a gene. Needed if you are selecting a single value (sample_size=1) to select a value that respects the allow_duplicate_genes parameter instead of selecting a value randomly. If None, then the gene value is selected randomly.
            -range_min (int, optional): The minimum value in the range from which a value is selected. It must be passed for generating the gene value randomly because we cannot decide whether it is the range for the initial population (init_range_low and init_range_high) or mutation (random_mutation_min_val and random_mutation_max_val).
            -range_max (int, optional): The maximum value in the range from which a value is selected. It must be passed for generating the gene value randomly because we cannot decide whether it is the range for the initial population (init_range_low and init_range_high) or mutation (random_mutation_min_val and random_mutation_max_val).
            -sample_size: The number of random values to generate/select and return. It tries to generate a number of values up to a maximum of sample_size. But it is not always guaranteed because the total number of values might not be enough or the random generator creates duplicate random values. For int data types, it could be None to keep all the values. For float data types, a None value returns only a single value.
            -step (int, optional): The step size for generating candidate values. Defaults to 1. Only used with genes of an integer data type.

        It returns,
            -A single numeric value if sample_size=1. Or
            -An array with number of values equal to sample_size if sample_size>1.
        """
        if self.gene_space is None:
            output = self.generate_gene_value_randomly(range_min=range_min,
                                                       range_max=range_max,
                                                       gene_value=gene_value,
                                                       gene_idx=gene_idx,
                                                       mutation_by_replacement=mutation_by_replacement,
                                                       sample_size=sample_size,
                                                       step=step)
        else:
            output = self.generate_gene_value_from_space(gene_value=gene_value,
                                                         gene_idx=gene_idx,
                                                         mutation_by_replacement=mutation_by_replacement,
                                                         solution=solution,
                                                         sample_size=sample_size)
        return output

    def get_valid_gene_constraint_values(self,
                                         range_min,
                                         range_max,
                                         gene_value,
                                         gene_idx,
                                         mutation_by_replacement,
                                         solution,
                                         sample_size=100,
                                         step=1):
        """
        Generate/select values for the gene that satisfy the constraint. The values could be generated randomly or from the gene space.
        The number of returned values is at its maximum equal to the sample_size parameter.
        It accepts:
            -range_min: The minimum value in the range from which a value is selected.
            -range_max: The maximum value in the range from which a value is selected.
            -gene_value: The original gene value before applying mutation.
            -gene_idx: The index of the gene in the solution.
            -mutation_by_replacement: A flag indicating whether mutation by replacement is enabled or not. The reason is to make this helper method usable while generating the initial population. In this case, mutation_by_replacement does not matter and should be considered False.
            -solution: The solution in which the gene exists.
            -sample_size: The number of values to generate or select. It tries to generate a number of values up to a maximum of sample_size. But it is not always guaranteed because the total number of values might not be enough or the random generator creates duplicate random values.
            -step (int, optional): The step size for generating candidate values. Defaults to 1. Only used with genes of an integer data type.

        It returns,
            -A single numeric value if sample_size=1. Or
            -An array with number of values equal to sample_size if sample_size>1. Or
            -None if no value found that satisfies the constraint.
        """

        # Either generate the values randomly or from the gene space.
        values = self.generate_gene_value(range_min=range_min,
                                          range_max=range_max,
                                          gene_value=gene_value,
                                          gene_idx=gene_idx,
                                          mutation_by_replacement=mutation_by_replacement,
                                          sample_size=sample_size,
                                          step=step)
        # It returns None if no value found that satisfies the constraint.
        values_filtered = self.filter_gene_values_by_constraint(values=values,
                                                                solution=solution,
                                                                gene_idx=gene_idx)
        return values_filtered
